# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2019-2021 Northwestern University.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio App RDM is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Common pytest fixtures and plugins."""

# Monkey patch Werkzeug 2.1
# Flask-Login uses the safe_str_cmp method which has been removed in Werkzeug
# 2.1. Flask-Login v0.6.0 (yet to be released at the time of writing) fixes the
# issue. Once we depend on Flask-Login v0.6.0 as the minimal version in
# Flask-Security-Invenio/Invenio-Accounts we can remove this patch again.
try:
    # Werkzeug <2.1
    from werkzeug import security

    security.safe_str_cmp
except AttributeError:
    # Werkzeug >=2.1
    import hmac

    from werkzeug import security

    security.safe_str_cmp = hmac.compare_digest

import shutil
import tempfile
from collections import namedtuple

import pytest
from flask_security import login_user
from flask_security.utils import hash_password
from invenio_access.models import ActionUsers
from invenio_access.permissions import system_identity
from invenio_access.proxies import current_access
from invenio_accounts.proxies import current_datastore
from invenio_accounts.testutils import login_user_via_session
from invenio_app.factory import create_app as _create_app
from invenio_db import db
from invenio_files_rest.models import Bucket, FileInstance, Location, ObjectVersion
from invenio_records_resources.proxies import current_service_registry
from invenio_vocabularies.contrib.subjects.api import Subject
from invenio_vocabularies.proxies import current_service as vocabulary_service
from invenio_vocabularies.records.api import Vocabulary
from invenio_vocabularies.records.models import VocabularyScheme


@pytest.fixture(scope="module")
def create_app(entry_points):
    """Creates a test app."""
    return _create_app


@pytest.fixture(scope="module")
def app_config(app_config):
    """Override pytest-invenio app_config fixture to disable CSRF check."""
    # Variable not used. We set it to silent warnings
    app_config["REST_CSRF_ENABLED"] = False

    return app_config


@pytest.fixture(scope="module")
def subjects_service(app):
    """Subjects service."""
    return current_service_registry.get("subjects")


pytest_plugins = ("celery.contrib.pytest",)


@pytest.fixture(scope="module")
def extra_entry_points():
    """Register extra entry point."""
    return {
        "invenio_base.blueprints": ["mock_module = mock_module.views:create_blueprint"],
    }


@pytest.fixture(scope="function")
def minimal_record(users):
    """Minimal record data as dict coming from the external world."""
    return {
        "access": {
            "record": "public",
            "files": "public",
        },
        "files": {"enabled": False},  # Most tests don't care about file upload
        "metadata": {
            "publication_date": "2020-06-01",
            "resource_type": {
                "id": "image-photo",
            },
            # Technically not required
            "creators": [
                {
                    "person_or_org": {
                        "type": "personal",
                        "name": "Doe, John",
                        "given_name": "John Doe",
                        "family_name": "Doe",
                    }
                }
            ],
            "title": "A Romans story",
        },
    }


@pytest.fixture()
def users(app, db):
    """Create users."""
    password = "123456"
    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        # create users
        hashed_password = hash_password(password)
        user1 = datastore.create_user(
            email="user1@test.com", password=hashed_password, active=True
        )
        user2 = datastore.create_user(
            email="user2@test.com", password=hashed_password, active=True
        )
        # Give role to administration-access
        db.session.add(ActionUsers(action="administration-access", user=user1))
    db.session.commit()
    return {
        "user1": user1,
        "user2": user2,
    }


@pytest.fixture()
def roles(app, db):
    """Create some roles."""
    with db.session.begin_nested():
        datastore = app.extensions["security"].datastore
        role1 = datastore.create_role(
            name="administration", description="administration role"
        )
        role2 = datastore.create_role(name="test", description="tests are coming")

    db.session.commit()
    return {"administration": role1, "test": role2}


@pytest.fixture()
def administration_user(users, roles):
    """Give administration rights to a user."""
    user = users["user1"]
    role = roles["administration"]
    current_datastore.add_role_to_user(user, role)
    action = current_access.actions["superuser-access"]
    db.session.add(ActionUsers.allow(action, user_id=user.id))

    return user


@pytest.fixture()
def client_with_login(client, users):
    """Log in a user to the client."""
    user = users["user1"]
    login_user(user, remember=True)
    login_user_via_session(client, email=user.email)
    return client


@pytest.fixture(scope="module")
def resource_type_type(app):
    """Resource type vocabulary type."""
    return vocabulary_service.create_type(system_identity, "resourcetypes", "rsrct")


@pytest.fixture(scope="module")
def resource_type_item(app, resource_type_type):
    """Resource type vocabulary record."""
    rst = vocabulary_service.create(
        system_identity,
        {
            "id": "image-photo",
            "icon": "chart bar outline",
            "props": {
                "csl": "graphic",
                "datacite_general": "Image",
                "datacite_type": "Photo",
                "eurepo": "info:eu-repo/semantic/image-photo",
                "openaire_resourceType": "25",
                "openaire_type": "dataset",
                "schema.org": "https://schema.org/Photograph",
                "subtype": "image-photo",
                "type": "image",
            },
            "title": {"en": "Photo"},
            "tags": ["depositable", "linkable"],
            "type": "resourcetypes",
        },
    )

    Vocabulary.index.refresh()

    return rst


@pytest.fixture(scope="module")
def languages_type(app):
    """Language vocabulary type."""
    return vocabulary_service.create_type(system_identity, "languages", "lng")


@pytest.fixture(scope="module")
def language_item(app, languages_type):
    """Language vocabulary record."""
    lang = vocabulary_service.create(
        system_identity,
        {
            "id": "eng",
            "props": {
                "alpha_2": "",
            },
            "title": {"en": "English"},
            "type": "languages",
        },
    )

    Vocabulary.index.refresh()

    return lang


@pytest.fixture
def subjects_mesh_scheme(app, db):
    """Subject Scheme for MeSH."""
    scheme = VocabularyScheme.create(
        id="MeSH",
        parent_id="subjects",
        name="Medical Subject Headings",
        uri="https://www.nlm.nih.gov/mesh/meshhome.html",
    )
    db.session.commit()
    return scheme


@pytest.fixture
def subject_item(app, subjects_mesh_scheme, subjects_service):
    """Subject vocabulary record."""
    subj = subjects_service.create(
        system_identity,
        {
            "id": "https://id.nlm.nih.gov/mesh/D000015",
            "scheme": "MeSH",
            "subject": "Abnormalities, Multiple",
        },
    )

    Subject.index.refresh()

    return subj


RunningApp = namedtuple(
    "RunningApp",
    ["app", "location", "resource_type_item", "language_item", "subject_item"],
)


@pytest.fixture
def running_app(app, location, resource_type_item, language_item, subject_item):
    """Fixture mimicking a running app."""
    return RunningApp(app, location, resource_type_item, language_item, subject_item)


@pytest.yield_fixture()
def dummy_location(db):
    """File system location."""
    tmppath = tempfile.mkdtemp()

    loc = Location(name="testloc", uri=tmppath, default=True)
    db.session.add(loc)
    db.session.commit()

    yield loc

    shutil.rmtree(tmppath)


@pytest.fixture
def invalid_file_instance(db, dummy_location):
    """Creates a file instance."""
    # Create a Bucket and ObjectVersion
    b1 = Bucket.create(location=dummy_location)
    with open("README.rst", "rb") as fp:
        obj = ObjectVersion.create(b1, "README.rst", stream=fp)
    db.session.commit()
    file_id = obj.file_id

    # Get FileInstance from file ID
    f = FileInstance.query.get(file_id)

    # Force an invalid checksum
    f.checksum = "invalid"
    f.verify_checksum()
    db.session.commit()

    # Retrieve the file instance (with updated last_check)
    f = FileInstance.query.get(file_id)

    return f



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2022 Graz University of Technology.
#
# Invenio-App-RDM is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Record migration fix script from InvenioRDM 8.0 to 9.0."""

from click import secho
from invenio_db import db
from invenio_rdm_records.records.api import RDMDraft, RDMRecord


def execute_upgrade_fix():
    """Execute the upgrade fix from InvenioRDM 8.0 to 9.0."""

    def update_parent(record):
        """Update parent schema and parent communities for older records."""
        new_parent_schema = "local://records/parent-v2.0.0.json"
        record.parent["$schema"] = new_parent_schema
        record.parent.setdefault("communities", {})

    def update_record(record):
        # skipping deleted records because can't be committed
        if record.is_deleted:
            return

        try:
            secho(f"Updating record : {record.pid.pid_value}", fg="yellow")

            update_parent(record)

            record.parent.commit()
            record.commit()

            secho(f"> Updated parent: {record.parent.pid.pid_value}", fg="green")
            secho(f"> Updated record: {record.pid.pid_value}\n", fg="green")
            return None
        except Exception as e:
            secho("> Error {}".format(repr(e)), fg="red")
            error = "Record {} failed to update".format(record.pid.pid_value)
            return error

    errors = []
    for record_metadata in RDMRecord.model_cls.query.all():
        record = RDMRecord(record_metadata.data, model=record_metadata)
        error = update_record(record)
        if error:
            errors.append(error)

    for draft_metadata in RDMDraft.model_cls.query.all():
        draft = RDMDraft(draft_metadata.data, model=draft_metadata)
        error = update_record(draft)
        if error:
            errors.append(error)

    success = not errors

    if success:
        secho("Committing to DB", nl=True)
        db.session.commit()
        secho(
            "Data migration completed, please rebuild the search indices now.",
            fg="green",
        )

    else:
        secho("Rollback", nl=True)
        db.session.rollback()
        secho(
            "Upgrade aborted due to the following errors:",
            fg="red",
            err=True,
        )

        for error in errors:
            secho(error, fg="red", err=True)

        msg = (
            "The changes have been rolled back. "
            "Please fix the above listed errors and try the upgrade again",
        )
        secho(msg, fg="yellow", err=True)


if __name__ == "__main__":
    execute_upgrade_fix()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2024 CERN.
# Copyright (C) 2019-2020 Northwestern University.
# Copyright (C)      2021 TU Wien.
# Copyright (C) 2023-2024 Graz University of Technology.
# Copyright (C)      2024 KTH Royal Institute of Technology.
#
# Invenio App RDM is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Routes for general pages provided by Invenio-App-RDM."""

from flask import Blueprint, current_app, flash, render_template, request
from flask_login import current_user
from invenio_db import db
from invenio_i18n import get_locale
from invenio_i18n import lazy_gettext as _
from invenio_pages.views import create_page_view
from invenio_users_resources.forms import NotificationsForm


def create_url_rule(rule, default_view_func):
    """Generate rule from string or tuple."""
    if isinstance(rule, tuple):
        path, view_func = rule

        return {"rule": path, "view_func": view_func}
    else:
        return {"rule": rule, "view_func": default_view_func}


#
# Registration
#
def create_blueprint(app):
    """Blueprint for the routes and resources provided by Invenio-App-RDM."""
    routes = app.config.get("APP_RDM_ROUTES")

    blueprint = Blueprint(
        "invenio_app_rdm",
        __name__,
        template_folder="templates",
        static_folder="static",
    )

    blueprint.add_url_rule(**create_url_rule(routes["index"], default_view_func=index))
    blueprint.add_url_rule(
        **create_url_rule(routes["robots"], default_view_func=robots)
    )
    blueprint.add_url_rule(
        **create_url_rule(routes["help_search"], default_view_func=help_search)
    )
    blueprint.add_url_rule(
        **create_url_rule(routes["help_statistics"], default_view_func=help_statistics)
    )
    blueprint.add_url_rule(
        **create_url_rule(routes["help_versioning"], default_view_func=help_versioning)
    )
    add_static_page_routes(blueprint, app)

    return blueprint


#
# Views
#
def index():
    """Frontpage."""
    return render_template(
        current_app.config["THEME_FRONTPAGE_TEMPLATE"],
        show_intro_section=current_app.config["THEME_SHOW_FRONTPAGE_INTRO_SECTION"],
    )


def robots():
    """Robots.txt."""
    return current_app.send_static_file("robots.txt")


def help_search():
    """Search help guide."""
    # Default to rendering english page if locale page not found.
    locale = get_locale()
    return render_template(
        [
            f"invenio_app_rdm/help/search.{locale}.html",
            "invenio_app_rdm/help/search.en.html",
        ]
    )


def help_statistics():
    """Statistics help guide."""
    # Default to rendering english page if locale page not found.
    locale = get_locale()
    return render_template(
        [
            f"invenio_app_rdm/help/statistics.{locale}.html",
            "invenio_app_rdm/help/statistics.en.html",
        ]
    )


def help_versioning():
    """DOI versioning help guide."""
    # Default to rendering english page if locale page not found.
    locale = get_locale()
    return render_template(
        [
            f"invenio_app_rdm/help/versioning.{locale}.html",
            "invenio_app_rdm/help/versioning.en.html",
        ]
    )


def notification_settings():
    """View for notification settings."""
    preferences_notifications_form = NotificationsForm(
        formdata=None, obj=current_user, prefix="preferences-notifications"
    )

    # Pick form
    form_name = request.form.get("submit", None)
    form = preferences_notifications_form if form_name else None

    # Process form
    if form:
        form.process(formdata=request.form)
        if form.validate_on_submit():
            handle_notifications_form(form)
            flash(_("Notification preferences were updated."), category="success")

    return render_template(
        "invenio_app_rdm/settings/notifications.html",
        preferences_notifications_form=preferences_notifications_form,
    )


def handle_notifications_form(form):
    """Handle notification preferences form."""
    form.populate_obj(current_user)
    db.session.add(current_user)
    current_app.extensions["security"].datastore.commit()


def add_static_page_routes(blueprint, app):
    """Add custom page routes to the blueprint from the app configuration."""
    for endpoint, path in app.config["APP_RDM_PAGES"].items():
        blueprint.add_url_rule(
            path, endpoint=endpoint, view_func=create_page_view(path)
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 CERN.
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-App-RDM is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Record migration script from InvenioRDM 11.0 to 12.0.

Disclaimer: This script is intended to be executed *only once*, namely when
upgrading from InvenioRDM 11.0 to 12.0!
If this script is executed at any other time, probably the best case scenario
is that nothing happens!


This script has been tested with following data:

- user
  - demo records of v11
  - demo communities of v11
  - cli created user of v11
  - ui created community (com_a)
  - ui created community private (com_b)
  - ui created record (rec_a.v1)
  - ui created v2 of record (rec_a.v1)
  - ui created record (rec_b.v1) added to community (com_a)
  - ui created draft (dra_a)
  - ui created draft (dra_b) added to community (com_b)
  - repository with records without managed doi
  - repository with records with managed doi
  - repository with records with managed doi without parent doi after migration
  - base vocabularies (no customized) usable after migration
  - record (rec_a.v1, rec_a.v2) findable after migration and rebuild of index

- administration
  - user panel list of users visible
  - drafts visible
  - records visible
"""
import sys

from click import secho
from flask import current_app
from invenio_access.permissions import system_identity
from invenio_communities.communities.records.api import Community
from invenio_communities.communities.records.systemfields.access import ReviewPolicyEnum
from invenio_db import db
from invenio_rdm_records.fixtures import PrioritizedVocabulariesFixtures
from invenio_rdm_records.proxies import current_rdm_records
from invenio_rdm_records.records.api import RDMDraft, RDMRecord


def execute_upgrade():
    """Execute the upgrade from InvenioRDM 11.0 to 12.0.0.

    Please read the disclaimer on this module before thinking about executing
    this function!
    """

    def migrate_review_policy(community_record):
        if community_record.is_deleted:
            return

        community_record["access"].setdefault(
            "review_policy", ReviewPolicyEnum.CLOSED.value
        )

    def update_parent(record):
        """Update parent schema and parent communities for older records."""
        new_parent_schema = "local://records/parent-v3.0.0.json"
        record.parent["$schema"] = new_parent_schema

        if (
            isinstance(record.parent["access"]["owned_by"], list)
            and len(record.parent["access"]["owned_by"]) > 0
        ):
            record.parent.access.owned_by = {
                "user": record.parent["access"]["owned_by"][0]["user"]
            }

        if "pids" not in record.parent:
            record.parent["pids"] = {}

            if (
                current_app.config["DATACITE_ENABLED"]
                and "doi" in current_app.config["RDM_PARENT_PERSISTENT_IDENTIFIERS"]
                and current_app.config["RDM_PARENT_PERSISTENT_IDENTIFIERS"]["doi"][
                    "is_enabled"
                ]
            ):
                pids = current_rdm_records.records_service.pids.parent_pid_manager.create_all(
                    record.parent, pids={}, schemes={"doi"}
                )
                current_rdm_records.records_service.pids.parent_pid_manager.reserve_all(
                    record.parent, pids
                )
                record.parent["pids"] = pids
                # Have to commit here otherwise register_or_update won't get
                # the above data
                record.parent.commit()

                if isinstance(record, RDMRecord):
                    current_rdm_records.records_service.pids.register_or_update(
                        id_=record["id"],
                        identity=system_identity,
                        scheme="doi",
                        parent=True,
                    )

    def update_record(record):
        # skipping deleted records because can't be committed
        if record.is_deleted:
            return

        try:
            secho(f"Updating record : {record.pid.pid_value}", fg="yellow")

            # otherwise the save would not work, due to new attributes
            # (media_files, parent_doi) used
            record["$schema"] = "local://records/record-v6.0.0.json"

            # Initialize media files as disabled if not any
            record.setdefault("media_files", {"enabled": False})
            if record.media_files.bucket is None:
                record.media_files.create_bucket()

            update_parent(record)

            record.commit()

            secho(f"> Updated parent: {record.parent.pid.pid_value}", fg="green")
            secho(f"> Updated record: {record.pid.pid_value}\n", fg="green")
            return None
        except Exception as e:
            secho(f"> Error {repr(e)}", fg="red")
            error = f"Record {record.pid.pid_value} failed to update"
            return error

    secho("Starting data migration...", fg="green")

    # upgrading vocabularies
    pvf = PrioritizedVocabulariesFixtures(system_identity)
    pvf.load()

    # Migrating communities
    communities = Community.model_cls.query.all()

    for community_data in communities:
        community = Community(community_data.data, model=community_data)

        # production data could have problems without it
        if community:
            migrate_review_policy(community)
            community.commit()

    # Migrating records and drafts
    errors = []
    for record_metadata in RDMRecord.model_cls.query.all():
        record = RDMRecord(record_metadata.data, model=record_metadata)
        error = update_record(record)

        if error:
            errors.append(error)

    for draft_metadata in RDMDraft.model_cls.query.all():
        draft = RDMDraft(draft_metadata.data, model=draft_metadata)
        error = update_record(draft)
        if error:
            errors.append(error)

    success = not errors

    if success:
        secho("Commiting to DB", nl=True)
        db.session.commit()
        secho(
            "Data migration completed, please rebuild the search indices now.",
            fg="green",
        )

    else:
        secho("Rollback", nl=True)
        db.session.rollback()
        secho(
            "Upgrade aborted due to the following errors:",
            fg="red",
            err=True,
        )

        for error in errors:
            secho(error, fg="red", err=True)

        msg = (
            "The changes have been rolled back. "
            "Please fix the above listed errors and try the upgrade again",
        )
        secho(msg, fg="yellow", err=True)

        sys.exit(1)


# if the script is executed on its own, perform the upgrade
if __name__ == "__main__":
    execute_upgrade()
